///--------------------------------------------------------------------------------------
// Scrolling Demo.c
//
// By Vern Jensen. Created in August of 1995
///--------------------------------------------------------------------------------------


#include <SWFPSReport.h>
#include <SWIncludes.h>
#include <SWGameUtils.h>
#include <SWStats.h>
#include <SWParticles.h>
#include <SWApplication.h>

#include "Scrolling Demo.h"


#define	kFullScreenWindow			true		// Try me full screen!
#define kWorldRectInset				0			// Make the SpriteWorld smaller?
#define	kInterlacedMode				false		// Skips every other line
#define kSyncToVBL					false		// Sync SpriteWorld to VBL?
#define kMaxFPS						30			// Set to 0 for unrestricted speed

#define kDoParticleExplosion		true		// Allows you to turn particle effects on/off	

#define kNumParticles				3500		// Max number of particles
#define kGravity		 SW_FLOAT2FIX(.8)		// Gravity applied to each particle
#define kExplosionLifeRemaining		60			// How long the particles last


#define	kSpriteMoveDelta			20			// Try 5, 10, 20, or 40
#define kDiamondSpace				8			// How far apart the diamonds are spaced
												// (try 1!)

#define kSpriteMoveDistance			80			// How far the sprite can move from
												// the center of the screen, in pixels.
												// Try making this value higher!

#define kStatsOffset				85			// Distance the numbers are from the right side
												// of the screen.

	// Number of ticks to wait before changing tile image; 0 = change every frame
#define kDiamondFrameRate			8			// How often diamond tiles change frames
#define kWallFrameRate				120			// How often wall changes frames

#define kBWPictResIDOffset			100
#define kTileWidth					40
#define kTileHeight					40
#define kTileMapRows				100
#define kTileMapCols				150

#define kStartRow					5			// Starting position of sprite
#define kStartCol					5			// in tile col and row


#define	kLeftArrowKey				0x7B
#define	kRightArrowKey				0x7C
#define	kDownArrowKey				0x7D
#define	kUpArrowKey					0x7E

#define	kLeftKeyPad					0x56
#define	kRightKeyPad				0x58
#define	kDownKeyPad					0x54
#define	kUpKeyPad					0x5B

#define kEscKey						0x35


#define kNoKey						0
#define kLeftKey					1
#define kUpKey						2
#define kRightKey					3
#define kDownKey					4


enum tileIDs
{
	kWallTile,
	kLastWallTile,
	kGrassTile,
	kBlackTile,
	kDiamondTile,
	kDiamondTile2,
	kLastDiamondTile,
	
	kTunnelTile1,
	kTunnelTile2,
	kTunnelTile3,
	kTunnelTile4,
	kTunnelTile5,
	kTunnelTile6,
	kWireTile1,
	kWireTile2,
	
	kMaxNumTiles
};



/***********/
/* Globals */
/***********/

SpriteWorldPtr		gSpriteWorldP;
SpriteLayerPtr		gSpriteLayerP, gNonScrollingSpriteLayerP;
TileMapStructPtr	gTileMapStructP;
TileMapPtr			gTileMap;
SpritePtr			gSimpleSpriteP, gStatsSpriteP, gDiamondMeterSpriteP, gMasterStatsSpriteP;
DrawProcPtr			gSpriteDrawProc;
WindowPtr			gWindowP;
Rect				gScreenMidRect;

long				gNumDiamondsInMap = 0;	// Number of diamonds in the TileMap
long				gNumDiamonds = 0;		// Number of diamonds sprite has collected

struct moveKeys		// Keeps track of which keys are up and which are down
{
	Boolean	up;
	Boolean	right;
	Boolean	down;
	Boolean	left;
} gKeys;


///--------------------------------------------------------------------------------------
// Main
///--------------------------------------------------------------------------------------

void	main( void )
{
	Initialize(kNumberOfMoreMastersCalls);

	if (SWHasSystem7())
	{
		AllowKeyUpEvents();	// Part of SWGameUtils.c
		SetCursor(*GetCursor(watchCursor));
		HideControlStrip();
		SWSetCleanUpFunction(MyCleanUpFunction);
		
		CreateSpriteWorld();
		SetUpTiling();
		CreateBallSprite();
		CreateDiamondMeterSprite();
		CreateStatsNumberSprite();
		
		SetCursor(&qd.arrow);
		HideCursor();
		
		SetUpAnimation();
		RunAnimation();
		ShutDown();
		
		RestoreControlStrip();	// Call this after using HideControlStrip before quitting
		RestoreEventMask();		// Call this after using AllowKeyUpEvents before quitting
	}
	else
	{
		CantRunOnThisMachine();
	}
}


///--------------------------------------------------------------------------------------
// CreateSpriteWorld
///--------------------------------------------------------------------------------------

void	CreateSpriteWorld( void )
{
	Rect		offscreenRect, worldRect, windRect;
	RgnHandle	mBarUpdateRgn;
	OSErr		err;
	
	gWindowP = GetNewCWindow(kWindowResID, NULL, (WindowPtr)-1L);
	
	if (gWindowP != NULL)
	{
		if (kFullScreenWindow == true)
		{
			SizeWindow(gWindowP, qd.screenBits.bounds.right, 
					qd.screenBits.bounds.bottom, false);
			MoveWindow(gWindowP, 0, 0, false);
		}
		else
		{
				// Center window in screen
			windRect = gWindowP->portRect;
			CenterRect(&windRect, &qd.screenBits.bounds);
			
				// Make sure window is aligned to long-word boundaries (for 8-bit mode)
			windRect.left = windRect.left>>2<<2;
			
			MoveWindow(gWindowP, windRect.left, windRect.top, false);
		}
		
		ShowWindow(gWindowP);
		SetPort(gWindowP);
		mBarUpdateRgn = SWHideMenuBar(gWindowP); // Must be done *after* showing window!
		EraseRgn(mBarUpdateRgn);
		
		if (kInterlacedMode)
			PaintRect(&gWindowP->portRect);	// Blacken window for Interlaced mode
	}
	else
		CantFindResource();
	
	
	err = SWEnterSpriteWorld();
	FatalError(err);
	
	
	worldRect = gWindowP->portRect;
	InsetRect(&worldRect, kWorldRectInset, kWorldRectInset);
	
	
		// Set size of offscreen area
	offscreenRect = worldRect;
	OffsetRect(&offscreenRect, -offscreenRect.left, -offscreenRect.top);
	

		// Make offscreen area evenly divisible by tile width & height
	if ( (offscreenRect.right/kTileWidth)*kTileWidth != offscreenRect.right)
		offscreenRect.right = (offscreenRect.right/kTileWidth)*kTileWidth + kTileWidth;
	
	if ( (offscreenRect.bottom/kTileHeight)*kTileHeight != offscreenRect.bottom)
		offscreenRect.bottom = (offscreenRect.bottom/kTileHeight)*kTileHeight + kTileHeight;
	
		// Create the scrolling sprite world
	err = SWCreateSpriteWorldFromWindow(&gSpriteWorldP, (CWindowPtr)gWindowP, 
			&worldRect, &offscreenRect, 0);
	FatalError(err);
	
		// Create the sprite layers
	err = SWCreateSpriteLayer(&gSpriteLayerP);
	FatalError(err);
	err = SWCreateSpriteLayer(&gNonScrollingSpriteLayerP);
	FatalError(err);
	
		// Add them to the world
	SWAddSpriteLayer(gSpriteWorldP, gSpriteLayerP);					// Bottom
	SWAddSpriteLayer(gSpriteWorldP, gNonScrollingSpriteLayerP);		// Top
	
	
	gSpriteDrawProc = SWStdSpriteDrawProc;
	
		// Determine what DrawProcs to use
	if (gSpriteWorldP->pixelDepth >= 8)		// 8-bit, 16-bit, or 32-bit, for 68k and PPC
	{
		gSpriteDrawProc = BlitPixieMaskDrawProc;
		SWSetSpriteWorldOffscreenDrawProc(gSpriteWorldP, BlitPixieRectDrawProc);
		SWSetDoubleRectDrawProc(gSpriteWorldP, BlitPixieDoubleRectDrawProc);
		SWSetPartialMaskDrawProc(gSpriteWorldP, BlitPixiePartialMaskDrawProc);
	}
	else if ( SW_68K )	// Use the 68k-only AllBit blitters for depths below 8-bits.
	{
		gSpriteDrawProc = BlitPixieAllBitMaskDrawProc;
		SWSetSpriteWorldScreenDrawProc(gSpriteWorldP, BlitPixieAllBitRectDrawProc);
		SWSetSpriteWorldOffscreenDrawProc(gSpriteWorldP, BlitPixieAllBitRectDrawProc);
		SWSetPartialMaskDrawProc(gSpriteWorldP, BlitPixieAllBitPartialMaskDrawProc);
	}
}


///--------------------------------------------------------------------------------------
// SetUpTiling
///--------------------------------------------------------------------------------------

void	SetUpTiling( void )	
{
	short		resIDOffset;
	short		row, col;
	OSErr		err;

	err = SWInitTiling(gSpriteWorldP, kTileHeight, kTileWidth, kMaxNumTiles);
	FatalError(err);
	
	err = SWCreateTileMap(&gTileMapStructP, kTileMapRows, kTileMapCols);
	FatalError(err);
	
	SWInstallTileMap(gSpriteWorldP, gTileMapStructP, 0);
	gTileMap = gTileMapStructP->tileMap;
	
	SWSetSpriteLayerUnderTileLayer(gSpriteLayerP, 0);
	

		// Determine whether to load B&W or color graphics
	if (gSpriteWorldP->pixelDepth <= 2)
		resIDOffset = kBWPictResIDOffset;
	else
		resIDOffset = 0;

		// Load first set of tiles
	err = SWLoadTilesFromPictResource(
		gSpriteWorldP, 
		kWallTile,				// startTileID 
		kLastDiamondTile,		// endTileID
		200 + resIDOffset,		// pictResID
		0,						// maskResID
		kNoMask,				// maskType
		kMaskIsNotPartialMask,	// partialMask?
		1,						// horizBorderWidth
		1);						// vertBorderHeight
	FatalError(err);
	
		// Load masked set of tiles
	err = SWLoadTilesFromPictResource(
		gSpriteWorldP, 
		kTunnelTile1,			// startTileID 
		kWireTile2,				// endTileID
		201 + resIDOffset,		// pictResID
		401,					// maskResID
		kFatMask,				// maskType
		kMaskIsPartialMask,		// partialMask?
		1,						// horizBorderWidth
		1);						// vertBorderHeight
	FatalError(err);

	
		// Set up tileMap
	for (row = 0; row < kTileMapRows; row++)
	{
		for (col = 0; col < kTileMapCols; col++)
		{
			if (row == 0 || col == 0 || row == kTileMapRows-1 || col == kTileMapCols-1)
				gTileMap[row][col] = kWallTile;
			else if (row > kTileMapRows / 2)
				gTileMap[row][col] = kWireTile1;
			else if ((row/kDiamondSpace)*kDiamondSpace == row && 
					 (col/kDiamondSpace)*kDiamondSpace == col)
			{
				gTileMap[row][col] = kDiamondTile;
				gNumDiamondsInMap++;
			}
			else
				gTileMap[row][col] = kGrassTile;
		}
	}
	

		// Add tunnel to tileMap
	row = kTileMapRows / 2;

	for (col = 1; col < kTileMapCols-1; col += 2)
	{
		gTileMap[row][col] = kTunnelTile1;
		gTileMap[row][col+1] = kTunnelTile2;
		gTileMap[row+1][col] = kTunnelTile4;
		gTileMap[row+1][col+1] = kTunnelTile5;
	}
}


///--------------------------------------------------------------------------------------
// CreateBallSprite
///--------------------------------------------------------------------------------------

void	CreateBallSprite( void )
{
	OSErr	err;
	
		// Create the ball sprite
	err = SWCreateSpriteFromCicnResource(gSpriteWorldP, &gSimpleSpriteP, NULL, 
			128, 1, kFatMask);	
	FatalError(err);
	
		// Set up the ball sprite
	SWAddSprite(gSpriteLayerP, gSimpleSpriteP);
	SWSetSpriteMoveProc(gSimpleSpriteP, KeySpriteMoveProc);
	SWSetSpriteLocation(gSimpleSpriteP, kStartCol * kTileWidth, kStartRow * kTileHeight);
	SWSetSpriteMoveDelta(gSimpleSpriteP, 0, 0);
	SWSetSpriteDrawProc(gSimpleSpriteP, gSpriteDrawProc);
}


///--------------------------------------------------------------------------------------
// CreateDiamondMeterSprite
///--------------------------------------------------------------------------------------

void	CreateDiamondMeterSprite( void )
{
	short		spriteResID;
	OSErr		err;
	
		// Calculate resource ID for diamond meter sprite
	if (gSpriteWorldP->pixelDepth <= 2)
		spriteResID = 302;
	else
		spriteResID = 202;
	
		// Create the diamond meter sprite
	err = SWCreateSpriteFromPictResource(gSpriteWorldP,
				&gDiamondMeterSpriteP, 
				NULL,			// pointer to memory for sprite
				spriteResID, 	// picture resource id
				402,			// mask resource id
				1, 				// max frames
				kFatMask);		// mask type			
	FatalError(err);
	
	SWAddSprite(gNonScrollingSpriteLayerP, gDiamondMeterSpriteP);
	SWSetSpriteDrawProc(gDiamondMeterSpriteP, gSpriteDrawProc);	
	SWSetSpriteMoveProc(gDiamondMeterSpriteP, DiamondMeterSpriteMoveProc);
	SWSetSpriteLocation(gDiamondMeterSpriteP, 10, 10);
	
		// Mark this layer as a non-scrolling layer
	SWSetLayerAsNonScrolling(gNonScrollingSpriteLayerP, true);
}


///--------------------------------------------------------------------------------------
// CreateStatsNumberSprite
///--------------------------------------------------------------------------------------

void	CreateStatsNumberSprite( void )
{
	OSErr		err;
	
		// Load gMasterStatsSpriteP
	err = SWCreateSpriteFromSinglePict(gSpriteWorldP, &gMasterStatsSpriteP, 
			NULL, 128, 128, 24, 0, kFatMask);
	FatalError(err);
	
	SWLockSprite(gMasterStatsSpriteP);
	
		// Create the gNumLivesStatsSpriteP
	gStatsSpriteP = CreateStatsSpriteClone(gMasterStatsSpriteP, 3, kRightJustify, false);
	err = SetUpStatsSprite(gStatsSpriteP, gNonScrollingSpriteLayerP, gSpriteDrawProc,
			gSpriteWorldP->windRect.right - kStatsOffset, 8, 0);
	FatalError(err);
	
	SWSetSpriteMoveProc(gStatsSpriteP, StatsSpriteMoveProc);
}


///--------------------------------------------------------------------------------------
// SetUpAnimation
///--------------------------------------------------------------------------------------

void	SetUpAnimation( void )
{
	Rect		moveBoundsRect;
	OSErr		err;
	
	SWLockSpriteWorld(gSpriteWorldP);
	
		// Set up data used by the SmoothScrollingWorldMoveProc
	gScreenMidRect = gSimpleSpriteP->curFrameP->frameRect;
	CenterRect( &gScreenMidRect, &gSpriteWorldP->backRect );
	
	SWSetSpriteWorldMaxFPS(gSpriteWorldP, kMaxFPS);
	SWSyncSpriteWorldToVBL(gSpriteWorldP, kSyncToVBL);
	SWSetCleanUpSpriteWorld(gSpriteWorldP);
	
	if (gSpriteWorldP->pixelDepth >= 8)
	{
			// Set up particle stuff
		err = InitParticles(kNumParticles, kGravity);
		FatalError(err);
		
		SWSetPostEraseCallBack(gSpriteWorldP, EraseParticlesScrollingOffscreen);
		SWSetPostDrawCallBack(gSpriteWorldP, DrawParticlesScrollingOffscreen);
	}
	
		// movement boundary = size of tileMap
	SetRect(&moveBoundsRect, 0,0, kTileMapCols * kTileWidth, kTileMapRows * kTileHeight);
	
	SWSetScrollingWorldMoveBounds(gSpriteWorldP, &moveBoundsRect);
	SWSetScrollingWorldMoveProc(gSpriteWorldP, SmoothScrollingWorldMoveProc, gSimpleSpriteP);
	
	SWSetTileChangeProc(gSpriteWorldP, TileChangeProc);
	
		// Move visScrollRect to starting sprite position
	SWMoveVisScrollRect(gSpriteWorldP, 
		SWGetSpriteHorizLoc(gSpriteWorldP->followSpriteP) - gSpriteWorldP->backRect.right/2,
		SWGetSpriteVertLoc(gSpriteWorldP->followSpriteP) - gSpriteWorldP->backRect.bottom/2);
	
		// Make sure CopyBits, if used, doesn't try to colorize things
	SWSetPortToWindow(gSpriteWorldP);
	ForeColor(blackColor);
	BackColor(whiteColor);
	
	if (kInterlacedMode)
	{
		SWSetFrameInterlacingMode(gSpriteWorldP->workFrameP, true, kSkipOddLines);
		SWSetFrameInterlacingMode(gSpriteWorldP->windowFrameP, true, kSkipOddLines);
		
		SWSetPortToWorkArea(gSpriteWorldP);
		PaintRect(&gSpriteWorldP->backRect);
	}
	
	SWDrawTilesInBackground(gSpriteWorldP);
	SWUpdateScrollingSpriteWorld(gSpriteWorldP, true);
}


///--------------------------------------------------------------------------------------
//  RunAnimation
///--------------------------------------------------------------------------------------

void	RunAnimation( void )
{
	unsigned long		frames;
		
	frames = 0;
	StartTimer();

	FatalError( SWStickyError() ); // Make sure no errors got past us during setup

	while (!Button())
	{
		SWProcessScrollingSpriteWorld(gSpriteWorldP);
		FatalError( SWStickyError() );	// Make sure no errors occurred during a MoveProc, etc.
		SWAnimateScrollingSpriteWorld(gSpriteWorldP);
		
		SetStatsSpriteNumber(gStatsSpriteP, gNumDiamonds);
		
		if (gSpriteWorldP->frameHasOccurred)
		{
			if (gSpriteWorldP->pixelDepth >= 8)
				UpdateParticlesInScrollingWindow(gSpriteWorldP);
			
			frames++;
		}
	}
	
	
	SWShowMenuBar(gWindowP);
	ShowResults(frames);
}


#pragma mark -
///--------------------------------------------------------------------------------------
//  ShutDown (clean up and dispose of the SpriteWorld)
///--------------------------------------------------------------------------------------

void	ShutDown( void )
{
	SWDisposeSprite(&gMasterStatsSpriteP);
	SWDisposeSpriteWorld(&gSpriteWorldP);
	SWExitSpriteWorld();
	
	FlushEvents(everyEvent, 0);
	InitCursor();
}


///--------------------------------------------------------------------------------------
//  MyCleanUpFunction - called if an error occurs, to clean up before quitting
///--------------------------------------------------------------------------------------

void	MyCleanUpFunction( void )
{
	SWShowMenuBar(gWindowP);
	
	RestoreControlStrip();
	RestoreEventMask();	
}


///--------------------------------------------------------------------------------------
//  TileChangeProc
///--------------------------------------------------------------------------------------

SW_FUNC void TileChangeProc(
	SpriteWorldPtr spriteWorldP)
{
	short			curImage;
	static short	wallDelay = 0, diamondDelay = 0;
	static short	oldTicks = 0;
	short			ticksPassed, ticks;
	
		// Initialize oldTicks the first time this function is called
	if (oldTicks == 0)
		oldTicks = TickCount();
	
	ticks = TickCount();
	ticksPassed = ticks - oldTicks;		// Number of ticks passed since last call
	oldTicks = ticks;
	
		// kWallTile
	wallDelay += ticksPassed;
	if (wallDelay >= kWallFrameRate)
	{
		curImage = spriteWorldP->curTileImage[kWallTile];
		if (curImage < kLastWallTile)
			curImage++;
		else
			curImage = kWallTile;
		
		SWChangeTileImage(spriteWorldP, kWallTile, curImage);
		wallDelay = 0;
	}
	
	
		// kDiamondTile
	diamondDelay += ticksPassed;
	if (diamondDelay >= kDiamondFrameRate)
	{
		curImage = spriteWorldP->curTileImage[kDiamondTile];
		if (curImage < kLastDiamondTile)
			curImage++;
		else
			curImage = kDiamondTile;
		
		SWChangeTileImage(spriteWorldP, kDiamondTile, curImage);
		diamondDelay = 0;
	}
}


///--------------------------------------------------------------------------------------
//  KeySpriteMoveProc
///--------------------------------------------------------------------------------------

SW_FUNC void KeySpriteMoveProc(SpritePtr srcSpriteP)
{
	short	row, col;
	short	rowDelta, colDelta;
	short	tile;
	
	UpdateKeys();	// Put the latest key values in the keys structure
	
	
	row = SWGetSpriteVertLoc(srcSpriteP) / kTileHeight;
	col = SWGetSpriteHorizLoc(srcSpriteP) / kTileWidth;

	if (row * kTileHeight == SWGetSpriteVertLoc(srcSpriteP) &&
		col * kTileWidth == SWGetSpriteHorizLoc(srcSpriteP))
	{
		tile = gTileMap[row][col];
		
			// Leave black trail behind sprite
		if (tile == kGrassTile || tile == kDiamondTile)
			SWDrawTile(gSpriteWorldP, 0, row, col, kBlackTile);
		else if (tile == kTunnelTile2 || tile == kTunnelTile5 || tile == kWireTile1)
			SWDrawTile(gSpriteWorldP, 0, row, col, tile+1);
			
		if (tile == kDiamondTile)
		{
			gNumDiamonds++;			// Increase number of diamonds sprite has collected
			UpdateDiamondMeter();	// Update meter
			
			if (kDoParticleExplosion && gSpriteWorldP->pixelDepth >= 8)
				DoParticleExplosionForTile(row, col, tile);
		}
		

		rowDelta = 0;
		colDelta = 0;
		
		if (gKeys.left)
			colDelta = -1;
		else if (gKeys.right)
			colDelta = 1;
		else if (gKeys.up)
			rowDelta = -1;
		else if (gKeys.down)
			rowDelta = 1;
		
				
		
		tile = gTileMap[row + rowDelta][col + colDelta];
		
		if (tile != kWallTile && tile != kTunnelTile1 && tile != kTunnelTile4)
		{
			srcSpriteP->vertMoveDelta = rowDelta * kSpriteMoveDelta;
			srcSpriteP->horizMoveDelta = colDelta * kSpriteMoveDelta;
		}
		else
		{
			srcSpriteP->vertMoveDelta = 0;
			srcSpriteP->horizMoveDelta = 0;
		}
	}
	
	SWOffsetSprite(srcSpriteP, srcSpriteP->horizMoveDelta, srcSpriteP->vertMoveDelta);
	srcSpriteP->needsToBeDrawn = true;	// Otherwise the particles will erase parts of our sprite when it is idle
}


///--------------------------------------------------------------------------------------
//  DiamondMeterSpriteMoveProc
///--------------------------------------------------------------------------------------

SW_FUNC void DiamondMeterSpriteMoveProc(SpritePtr srcSpriteP)
{
	SWMoveSprite(srcSpriteP, 
		10 + gSpriteWorldP->visScrollRect.left, 
		10 + gSpriteWorldP->visScrollRect.top);
}


///--------------------------------------------------------------------------------------
//  StatsSpriteMoveProc
///--------------------------------------------------------------------------------------

SW_FUNC void StatsSpriteMoveProc(SpritePtr srcSpriteP)
{
	SWMoveSprite(srcSpriteP, 
		gSpriteWorldP->windRect.right - kStatsOffset + gSpriteWorldP->visScrollRect.left, 
		8 + gSpriteWorldP->visScrollRect.top);
}


///--------------------------------------------------------------------------------------
//  UpdateDiamondMeter - change the meter Sprite's image and mask
///--------------------------------------------------------------------------------------

SW_FUNC void UpdateDiamondMeter( void )
{
	double	percent;
	Rect	meterRect;
	
	percent = (double)gNumDiamonds / gNumDiamondsInMap;
	
		// 108 = length of meter; 17 = offset from start of meter
	SetRect(&meterRect, 17, 2, (108 * percent) + 17, 14);
	
	
		// Set port to our sprite's framePort GWorld
	SetGWorld(gDiamondMeterSpriteP->curFrameP->framePort, nil);
	
	ForeColor(magentaColor);
	PaintRect(&meterRect);
	
		// IMPORTANT: Set the color back when done! (In case CopyBits is used later)
	ForeColor(blackColor);
	

		// Update the pixel mask directly if we are using a pixelMask.
	if (gDiamondMeterSpriteP->frameDrawProc != SWStdSpriteDrawProc)
	{
			// Set port to our sprite's pixel mask GWorld
		SetGWorld(gDiamondMeterSpriteP->curFrameP->maskPort, nil);
		
			// Mask image is inverted when in 8-bit or less
		if (gSpriteWorldP->pixelDepth <= 8)
			ForeColor(whiteColor);
		else
			ForeColor(blackColor);
		
		PaintRect(&meterRect);
		ForeColor(blackColor);
	}
	else
	{
			// Otherwise, use SWUpdateFrameMasks to update the region mask.
		SWUpdateFrameMasks(gDiamondMeterSpriteP->curFrameP);
	}
	
	
		// Set sprite to be redrawn, since we've changed its image
	gDiamondMeterSpriteP->needsToBeDrawn = true;
}


///--------------------------------------------------------------------------------------
//  DoParticleExplosionForTile - turns each non-black pixel of a tile into a particle
///--------------------------------------------------------------------------------------

void DoParticleExplosionForTile(
	short tileRow, 
	short tileCol, 
	short tileID)
{	
	FramePtr		tileFrameP;
	short			row, col, pixelRow, pixelCol;
	unsigned long	color, black;
	float			hDelta, vDelta;
	short			halfWidth, halfHeight;
	RGBColor		blackRGB = {0,0,0};
	
	black = Color2Index( &blackRGB );
	
		// Get the FrameP for the tile at tileRow and tileCol
	tileFrameP = gSpriteWorldP->tileFrameArray[ gSpriteWorldP->curTileImage[tileID] ];
	SWSetPortToFrame(tileFrameP);
	
	pixelRow = tileRow * gSpriteWorldP->tileHeight;
	pixelCol = tileCol * gSpriteWorldP->tileWidth;
	
	halfWidth = (tileFrameP->frameRect.right - tileFrameP->frameRect.left) / 2;
	halfHeight = (tileFrameP->frameRect.bottom - tileFrameP->frameRect.top) / 2;
	
	for (row = tileFrameP->frameRect.top; row < tileFrameP->frameRect.bottom; row++)
		for (col = tileFrameP->frameRect.left; col < tileFrameP->frameRect.right; col++)
		{
			color = SWGetPixel(tileFrameP, col, row);
			
			if (color != black)
			{
				hDelta = (float)((col - tileFrameP->frameRect.left) - halfWidth) / 6;
				vDelta = (float)((row - tileFrameP->frameRect.top) - halfHeight) / 6 - 8;
				
				NewParticle( 
					color, 
					SW_INT2FIX(pixelCol + col - tileFrameP->frameRect.left),
					SW_INT2FIX(pixelRow + row - tileFrameP->frameRect.top),
					SW_FLOAT2FIX(hDelta),
					SW_FLOAT2FIX(vDelta),
					kExplosionLifeRemaining);
			}
		}
}


///--------------------------------------------------------------------------------------
//  SmoothScrollingWorldMoveProc - our scrolling WorldMoveProc
///--------------------------------------------------------------------------------------

SW_FUNC void SmoothScrollingWorldMoveProc(
	SpriteWorldPtr spriteWorldP,
	SpritePtr followSpriteP)
{	
	short	screenMidRectTop, screenMidRectLeft;
	
	screenMidRectTop = gScreenMidRect.top + spriteWorldP->visScrollRect.top;
	screenMidRectLeft = gScreenMidRect.left + spriteWorldP->visScrollRect.left;
	
	
	spriteWorldP->horizScrollDelta = (kSpriteMoveDelta * 
		(SWGetSpriteHorizLoc(followSpriteP) - screenMidRectLeft) ) / kSpriteMoveDistance;
		
		
	
	spriteWorldP->vertScrollDelta = (kSpriteMoveDelta * 
		(SWGetSpriteVertLoc(followSpriteP) - screenMidRectTop) ) / kSpriteMoveDistance;
	
	if (kInterlacedMode)
		spriteWorldP->vertScrollDelta = spriteWorldP->vertScrollDelta>>1<<1;
}


///--------------------------------------------------------------------------------------
//  UpdateKeys (Put the latest key values in the keys structure)
///--------------------------------------------------------------------------------------

void	UpdateKeys( void )
{
	EventRecord		event;
	short			theKey;
	Boolean			isDown;
	
	while ( GetOSEvent( (keyUpMask | keyDownMask), &event ) )
	{
		theKey = (event.message & keyCodeMask) >> 8;
		isDown = (event.what != keyUp);
		
		if ( (theKey == kLeftArrowKey) || (theKey == kLeftKeyPad) )
			gKeys.left = isDown;
		else if ( (theKey == kRightArrowKey) || (theKey == kRightKeyPad) )
			gKeys.right = isDown;
		else if ( (theKey == kDownArrowKey) || (theKey == kDownKeyPad) )
			gKeys.down = isDown;
		else if ( (theKey == kUpArrowKey) || (theKey == kUpKeyPad) )
			gKeys.up = isDown;
	}
}
