///--------------------------------------------------------------------------------------
// SpriteTest.h
///--------------------------------------------------------------------------------------
 

#ifndef __SPRITETEST__
#define __SPRITETEST__

#ifndef __MENUS__
#include <Menus.h>
#endif


enum
{
	kBaseResID = 128,
	kGlobeHeight = 48,
	kGlobeBorderWidth = 1,
	kNumberOfGlobeSprites = 10,
	kNumberOfGlobeFrames = 30,
	kGlobeSpriteMoveTime = 0,
	kGlobeSpriteFrameTime = 20,
	kTwoHeight = 50,
	kTwoBorderWidth = 1,
	kNumberOfTwoFrames = 10,
	kBackDropPixPatID = 128,
	kPerformanceAlertID = 200,
	kSetUpTestResID = 400,
	kCollisionProcResID = 401,
	kTestTime = (30 * 60)		// 30 seconds per test
};

typedef enum
{ 
	kCopyBitsProc	= 0, 
	kBlitPixieProc	= 1, 

	kRLEProc		= 2, 	// <-- sprite procs
	kCompiledProc	= 3
}ProcType;

typedef enum
{ 
	kRectCollisionProc		= 0, 
	kRadiusCollisionProc	= 1, 
	kPixelCollisionProc		= 2,
	kRegionCollisionProc	= 3,
	kRLECollisionProc		= 4
}CollType;

enum
{
	kCopyBitsSpriteButton = 4,
	kBlitPixieSpriteButton,
	kRLESpriteButton,
	kCompiledSpriteButton,
	kCopyBitsOffscreenButton = 9,
	kBlitPixieOffscreenButton,
	kCopyBitsScreenButton = 12,
	kBlitPixieScreenButton,
	kRunTestNowCheckBox,
	kInterlacedCheckBox,
	kSkipOddLinesRadioButton,
	kSkipEvenLinesRadioButton,
	kClearBackgroundCheckBox,
	kSetClearColorButton,
	kColorSpritesCheckBox,
	kSetSpriteColorButton
};

enum
{
	kCollisionDetectionCheckbox = 3,

	kRectCollisionButton		= 4, 
	kRadiusCollisionButton		= 5, 
	kPixelCollisionButton		= 6,
	kRegionCollisionButton		= 7,
	kRLECollisionButton			= 8

};

typedef struct SpriteTestRec SpriteTestRec;
typedef SpriteTestRec *SpriteTestPtr, **SpriteTestHdl;


struct SpriteTestRec
{
	SpriteWorldPtr spriteWorldP;
	SpriteLayerPtr globeSpriteLayerP;
	SpritePtr globeSpriteArray[kNumberOfGlobeSprites];
	SpritePtr titleSpriteP;
	SpritePtr twoSpriteP;
};
		

OSErr SetupSWStuff(
	SpriteTestPtr* spriteTestP,
	CWindowPtr srcWindowP);

void DisposeSWStuff(
	SpriteTestPtr spriteTestP);

void SetupSpriteWorldElements(
	SpriteTestPtr spriteTestP);

void SetDefaultTestDrawProcs( short depth );

void SetupGlobeSprite(
	SpritePtr testSpriteP,
	Rect *moveBoundsRect,
	short horizLocation,
	short vertLocation);

void SpriteTestIdle(
	SpriteTestPtr spriteTestP);

void UpdateSpriteTest(
	SpriteTestPtr spriteTestP,
	WindowPtr updateWindowP);

void HandleCreateSpriteCommand(
	SpriteTestPtr spriteTestP);

void HandleSpriteTestTitleCommand(
	SpriteTestPtr spriteTestP);

void HandleBouncingBallsCommand(
	SpriteTestPtr spriteTestP);

void RemoveClickedSprite(
	SpriteTestPtr spriteTestP);

void SetUpTestDialog(
	SpriteTestPtr spriteTestP);

void SetSetUpTestButtons(
	DialogPtr theDialog);

void CollisionProcDialog(
	SpriteTestPtr spriteTestP);

void SetCollisionProcButtons(
	DialogPtr theDialog );

void RunAllTests(
	SpriteTestPtr spriteTestP);
	
void RunTheTest(
	SpriteTestPtr spriteTestP);
	
void RestoreFromTest(
	WindowPtr testWindowP,
	SpriteTestPtr spriteTestP);
	
void DisplayPerformance(
	long frames,
	float seconds);
	
SW_FUNC void GlobeSpriteMoveProc(SpritePtr globeSpriteP);

SW_FUNC void RectBounceCollideProc(
	SpritePtr srcSpriteP,
	SpritePtr dstSpriteP,
	Rect* sectRect);

SW_FUNC void RLEBounceCollideProc(
	SpritePtr srcSpriteP,
	SpritePtr dstSpriteP,
	Rect* sectRect);

SW_FUNC void PixelBounceCollideProc(
	SpritePtr srcSpriteP,
	SpritePtr dstSpriteP,
	Rect* sectRect);
	
SW_FUNC void RegionBounceCollideProc(
	SpritePtr srcSpriteP,
	SpritePtr dstSpriteP,
	Rect* sectRect);

SW_FUNC void RadiusBounceCollideProc(
	SpritePtr srcSpriteP,
	SpritePtr dstSpriteP,
	Rect* sectRect);

void BounceGlobeOffGlobe(
	SpritePtr srcSpriteP,
	SpritePtr dstSpriteP);

void BounceGlobeOffTitle(
	SpritePtr titleSpriteP,
	SpritePtr globeSpriteP,
	Rect* sectRect);
	

#endif	/* __SPRITETEST__ */