#------------------------------------------------------------------------------
#
# Build script for separate MPW assembler files from inline CW assembly
#
# NOTE: Needs CodeWarrior MPW to build.
#
# Written by Anders F Bjrklund <afb@algonet.se>
#
#------------------------------------------------------------------------------

Prefix          = 

Headers         = -i "::Headers:"
Sources         =    "::Sources:"

obj             = :
asm68K          = "::Assembler:68k:"
asmPPC          = "::Assembler:PPC:"
asmX86          = "::Assembler:X86:"

IncludesFolders	= {Prefix} {Headers}
CCOptions       = -EOL -w noimplicitconv -d GENERATINGASM
CC68KOptions    = {CCOptions} -mbg off -mc68020
CCPPCOptions    = {CCOptions}

DumpOptions     =  -h -noheader -nodata -nonames -sym off

#------------------------------------------------------------------------------

all					68K PPC

68K					{asm68K}AsmGlobal.a 
					{asm68K}BlitPixieAllBit.a 
					{asm68K}BlitPixieClear.a 
					{asm68K}BlitPixieDoubled.a 
					{asm68K}BlitPixieDoubleRects.a 
					{asm68K}BlitPixieFlip.a 
					{asm68K}BlitPixieMask.a 
					{asm68K}BlitPixieMaskCollision.a 
					{asm68K}BlitPixieRect.a 
					{asm68K}BlitPixieRLE.a 
					{asm68K}BlitPixieRLECollision.a 
					{asm68K}BlitPixieUtils.a 
					{asm68K}BlitPixieBlitClear.a 
					{asm68K}BlitPixieBlitErase.a 
					{asm68K}BlitPixieDirtyRects.a 
					#
PPC					{asmPPC}AsmGlobal.s 
					{asmPPC}BlitPixieClear.s 
					{asmPPC}BlitPixieDoubled.s 
					{asmPPC}BlitPixieDoubleRects.s 
					{asmPPC}BlitPixieFlip.s 
					{asmPPC}BlitPixieMask.s 
					{asmPPC}BlitPixieMaskCollision.s 
					{asmPPC}BlitPixieRect.s 
					{asmPPC}BlitPixieRLE.s 
					{asmPPC}BlitPixieRLECollision.s 
					{asmPPC}BlitPixieUtils.s 
					{asmPPC}BlitPixieBlitClear.s 
					{asmPPC}BlitPixieBlitErase.s 
					{asmPPC}BlitPixieDirtyRects.s 
					#
#X86				{asmX86}

clean			
	Echo "# Removing generated asm"
	Delete -y {asm68K} {asmPPC}

#------------------------------------------------------------------------------

{obj}			 {Sources}
{asm68K}		 {obj}
{asmPPC}		 {obj}
#{asmX86}		 {obj}

.o					.c
	MWC68K   	{CC68KOptions} {IncludesFolders} {DepDir}{Default}.c -o {obj}{Default}.o

.x					.c
	MWCPPC		{CCPPCOptions} {IncludesFolders} {DepDir}{Default}.c -o {obj}{Default}.x

#.obj				.c
#	MWCX86		{CCPPCOptions} {IncludesFolders} {DepDir}{Default}.c -o {obj}{Default}.obj
	
.a					.o
	Echo		"# Generating 68k asm for "{Default}
	MWDump68K	{DumpOptions} {obj}{Default}.o |
	(StreamEdit	-set ASM_FUNCTION={Default} -s StreamScript68k >{TargDir}{Default}.a )
	Delete	    {obj}{Default}.o

.a					.a.in
	Echo		"# Generating 68k asm for "{Default}
	Catenate    {DepDir}{Default}.a.in >{TargDir}{Default}.a

.s					.x
	Echo		"# Generating PPC asm for "{Default}
	MWDumpPPC	{DumpOptions} {obj}{Default}.x |
	(StreamEdit	-set ASM_FUNCTION={Default} -s StreamScriptPPC >{TargDir}{Default}.s )
	Delete	    {obj}{Default}.x

.s					.s.in
	Echo		"# Generating PPC asm for "{Default}
	Catenate    {DepDir}{Default}.s.in >{TargDir}{Default}.s

#.asm				.obj
#	Echo		"# Generating X86 asm for "{Default}
	
#------------------------------------------------------------------------------

	
	
