///--------------------------------------------------------------------------------------
//	BlitPixieColorkey.c
//
//	written by Anders F Bjrklund <afb@algonet.se>
//	1999 afb.
///--------------------------------------------------------------------------------------

#ifndef __BLITPIXIE__
#include "BlitPixieHeader.h"
#endif

#include "BlitPixieAsm.h"

#if macintosh

#define kAlphaMask16	0x7FFF			// masking value to remove the alpha channel
#define kAlphaMask32	0x00FFFFFFL

#else

#define kAlphaMask16	0xFFFF			// compare with entire pixel, or change this
#define kAlphaMask32	0xFFFFFFFFL

#endif

#ifndef GENERATINGASM // do not include for asm file generation

///--------------------------------------------------------------------------------------
//	BlitPixieColorkey - since this routine is not the fastest way to do it, there is no asm.
//                      compare with BlitPixieMask, or BlitPixieRLE for other ways of masking.
///--------------------------------------------------------------------------------------

void BlitPixieColorKey8Bit(
	unsigned char *src,		
	unsigned char *dst,			
	unsigned long srcOffset,		
	unsigned long dstOffset,
	unsigned short width,		
	unsigned short height,	
	unsigned char colorKey )
{
	int		x;
	
	srcOffset -= width;
	dstOffset -= width;

	while ( height-- )
	{
		for (x = 0; x < width; x++)
		{
			if ( *src != colorKey)
				*dst = *src;
		
			src++;
			dst++;
		}
		
		src = (unsigned char *) ( (char *) src + srcOffset );
		dst = (unsigned char *) ( (char *) dst + dstOffset );
	}
}


void BlitPixieColorKey16Bit(
	unsigned short *src,		
	unsigned short *dst,			
	unsigned long srcOffset,		
	unsigned long dstOffset,
	unsigned short width,		
	unsigned short height,	
	unsigned short colorKey )
{
	int		x;
	
	srcOffset -= (width + width);
	dstOffset -= (width + width);
	
	colorKey &= kAlphaMask16;
	while ( height-- )
	{
		for (x = 0; x < width; x++)
		{
			if ( (*src & kAlphaMask16) != colorKey)
				*dst = *src;
		
			src++;
			dst++;
		}
		
		src = (unsigned short *) ( (char *) src + srcOffset );
		dst = (unsigned short *) ( (char *) dst + dstOffset );
	}
}


void BlitPixieColorKey32Bit(
	unsigned long *src,		
	unsigned long *dst,			
	unsigned long srcOffset,		
	unsigned long dstOffset,
	unsigned short width,		
	unsigned short height,	
	unsigned long colorKey )
{
	int		x;
	
	srcOffset -= (width << 2);
	dstOffset -= (width << 2);
	
	colorKey &= kAlphaMask32;
	while ( height-- )
	{
		for (x = 0; x < width; x++)
		{
			if ( (*src & kAlphaMask32) != colorKey)
				*dst++ = *src;
		
			src++;
			dst++;
		}

		src = (unsigned long *) ( (char *) src + srcOffset );
		dst = (unsigned long *) ( (char *) dst + dstOffset );
	}
}

#endif
